from typing import Any
from dataclasses import dataclass, field


@dataclass(slots=True)
class USBCameraConfig:
    device_index: int = field(
        default=0,
        metadata={
            "label": "Camera Device Index",
            "group": "Device",
            "ui": "number",
            "min": 0,
            "max": 16,
        },
    )

    format: str = field(
        default="MJPG",
        metadata={
            "label": "Pixel Format",
            "group": "Video",
            "ui": "select",
        },
    )

    width: int = field(
        default=1280,
        metadata={
            "label": "Frame Width",
            "group": "Video",
            "ui": "number",
            "min": 160,
            "max": 3840,
        },
    )

    height: int = field(
        default=720,
        metadata={
            "label": "Frame Height",
            "group": "Video",
            "ui": "number",
            "min": 120,
            "max": 2160,
        },
    )

    fps: int = field(
        default=30,
        metadata={
            "label": "Frames Per Second",
            "group": "Video",
            "ui": "number",
            "min": 1,
            "max": 120,
        },
    )

    auto_exposure: bool = field(
        default=True,
        metadata={
            "label": "Auto Exposure",
            "group": "Exposure",
        },
    )

    exposure: float = field(
        default=-6.0,
        metadata={
            "label": "Exposure",
            "group": "Exposure",
            "ui": "number",
            "min": -13.0,
            "max": -1.0,
            "step": 0.1,
        },
    )

    gain: float = field(
        default=1.0,
        metadata={
            "label": "Gain",
            "group": "Exposure",
            "help": "Amplification applied to the sensor signal",
            "ui": "number",
            "min": 0,
            "max": 16,
            "step": 0.1,
        },
    )

    auto_focus: bool = field(
        default=True,
        metadata={
            "label": "Auto Focus",
            "group": "Focus",
        },
    )

    focus: float = field(
        default=0.0,
        metadata={
            "label": "Focus",
            "group": "Focus",
            "ui": "number",
            "min": 0,
            "max": 255,
        },
    )

    auto_white_balance: bool = field(
        default=True,
        metadata={
            "label": "Auto White Balance",
            "group": "White Balance",
        },
    )

    white_balance_temperature: float = field(
        default=4500.0,
        metadata={
            "label": "White Balance Temperature (K)",
            "group": "White Balance",
            "ui": "number",
            "min": 2800,
            "max": 6500,
            "step": 10,
        },
    )

    @classmethod
    def from_dict(cls, data: dict[str, Any]) -> "USBCameraConfig":
        return cls(
            device_index=data.get("device_index", 0),
            format=data.get("format", "MJPG"),
            width=data.get("width", 1280),
            height=data.get("height", 720),
            fps=data.get("fps", 30),
            auto_exposure=data.get("auto_exposure", True),
            exposure=data.get("exposure", -6.0),
            gain=data.get("gain", 1.0),
            auto_focus=data.get("auto_focus", True),
            focus=data.get("focus", 0.0),
            auto_white_balance=data.get("auto_white_balance", True),
            white_balance_temperature=data.get("white_balance_temperature", 4500.0),
        )



from typing import Any
from dataclasses import dataclass, field


@dataclass(slots=True)
class W1BusConfig:
    base_path: str = field(
        default="/sys/bus/w1/devices",
        metadata={
            "label": "Base Path",
            "group": "Connection",
            "ui": "path",
        },
    )

    @classmethod
    def from_dict(cls, data: dict[str, Any]) -> "W1BusConfig":
        return cls(base_path=data.get("base_path", "/sys/bus/w1/devices"))



from bussdcc.device.definition import DeviceDefinition

from .config import DS18B20Config
from .driver import DS18B20

definition = DeviceDefinition(
    config_class=DS18B20Config,
    driver_class=DS18B20,
)



from bussdcc.device.definition import DeviceDefinition

from .config import NAU7802Config
from .driver import NAU7802

definition = DeviceDefinition(
    config_class=NAU7802Config,
    driver_class=NAU7802,
)



from typing import Any
from dataclasses import dataclass, field


@dataclass(slots=True)
class NAU7802CalibrationConfig:
    offset: int = field(
        default=0,
        metadata={
            "label": "Offset",
            "group": "Calibration",
            "ui": "number",
        },
    )

    scale: float = field(
        default=1.0,
        metadata={
            "label": "Scale Factor",
            "group": "Calibration",
            "ui": "number",
            "step": 0.0001,
        },
    )

    @classmethod
    def from_dict(cls, data: dict[str, Any]) -> "NAU7802CalibrationConfig":
        return cls(
            offset=data.get("offset", 0),
            scale=data.get("scale", 1.0),
        )


@dataclass(slots=True)
class NAU7802ChannelConfig:
    calibration: NAU7802CalibrationConfig = field(metadata={"group": "Channels"})

    @classmethod
    def from_dict(cls, data: dict[str, Any]) -> "NAU7802ChannelConfig":
        return cls(
            calibration=NAU7802CalibrationConfig.from_dict(data.get("calibration", {}))
        )


@dataclass(slots=True)
class NAU7802Config:
    bus_id: str = field(
        metadata={
            "label": "I²C Bus",
            "group": "Connection",
            "ui": "bus",
        }
    )

    addr: int = field(
        default=0x2A,
        metadata={
            "label": "I²C Address",
            "group": "Connection",
            "ui": "hex",
            "min": 0,
            "max": 127,
        },
    )

    channels: dict[str, NAU7802ChannelConfig] = field(
        default_factory=dict,
        metadata={"group": "Channels"},
    )

    @classmethod
    def from_dict(cls, data: dict[str, Any]) -> "NAU7802Config":
        channels = {
            str(k): NAU7802ChannelConfig.from_dict(v)
            for k, v in data.get("channels", {}).items()
        }

        return cls(
            bus_id=data["bus_id"],
            addr=data.get("addr", 0x2A),
            channels=channels,
        )



from typing import Optional, Any
from pathlib import Path

from bussdcc.device import Device

from ...bus.w1 import W1Bus

from .config import DS18B20Config


class DS18B20(Device[DS18B20Config]):
    kind = "temperature"

    def __init__(self, *, id: str, config: DS18B20Config):
        super().__init__(id=id, config=config)
        self._device_path: Path | None = None

    def connect(self) -> None:
        if not self.ctx:
            raise RuntimeError("Device not attached")

        bus = self.ctx.runtime.devices.get(self.config.bus_id)
        if not isinstance(bus, W1Bus):
            raise RuntimeError("Failed to find W1Bus")

        self._device_path = bus.device_path(self.config.device_id)
        if not self._device_path.exists():
            raise RuntimeError("Failed to find device")

    def read(self) -> float | None:
        if not self._device_path:
            return None

        temp_file = self._device_path / "temperature"

        try:
            value = temp_file.read_text().strip()
        except FileNotFoundError as e:
            self.set_offline(e)
            return None

        try:
            temp = float(value) / 1000
        except ValueError as e:
            self.set_offline(e)
            return None

        self.set_online()
        return temp



from typing import Any
from dataclasses import dataclass, field


@dataclass(slots=True)
class PumpConfig:
    pin: int = field(
        metadata={
            "label": "GPIO Output Pin",
            "group": "Hardware",
            "ui": "number",
            "min": 2,
            "max": 27,
        },
    )

    @classmethod
    def from_dict(cls, data: dict[str, Any]) -> "PumpConfig":
        return cls(pin=data["pin"])



from typing import Iterable, Optional, Any
from pathlib import Path

from bussdcc.device import Device

from .config import W1BusConfig


class W1Bus(Device[W1BusConfig]):
    kind = "bus"

    def __init__(self, *, id: str, config: W1BusConfig):
        super().__init__(id=id, config=config)
        self.base_path = Path(config.base_path)

    def connect(self) -> None:
        if not self.base_path.exists():
            raise RuntimeError("1-Wire bus not available")

    def disconnect(self) -> None:
        pass

    def discover_devices(self) -> Iterable[str]:
        return [
            p.name
            for p in self.base_path.iterdir()
            if p.is_dir() and p.name.startswith("28-")
        ]

    def device_path(self, device_id: str) -> Path:
        return Path(self.base_path / device_id)



from typing import Literal, Optional, Any
import threading
import time

from bussdcc.device import Device
from bussdcc.message import DeviceFailed
from nau7802 import NAU7802 as _NAU7802
from typed_registers import SMBusRegisterBus

from ...bus.i2c import I2CBus

from .config import NAU7802Config


class NAU7802(Device[NAU7802Config]):
    kind = "adc"

    def __init__(self, *, id: str, config: NAU7802Config):
        super().__init__(id=id, config=config)
        self._current_channel: int | None = None
        self._lock = threading.Lock()

    def connect(self) -> None:
        if not self.ctx:
            raise RuntimeError("Device not attached")

        runtime = self.ctx.runtime
        bus = runtime.devices.get(self.config.bus_id)

        if not isinstance(bus, I2CBus):
            raise RuntimeError("NAU7802 requires an I2CBus")

        self.bus: I2CBus = bus

        protocol: SMBusRegisterBus | None = bus.protocol()

        if protocol:
            self.device = _NAU7802(protocol, self.config.addr)

        with self._lock:
            self.device.initialize()
            self.device.set_crs(3)
            self.device.set_gain(7)

    def disconnect(self) -> None:
        pass

    def _wait_ready(self) -> bool:
        if not self.ctx:
            return False

        timeout = self.ctx.clock.monotonic() + 0.15  # at a minimum we expect 10 SPS
        while not self.device.cycle_ready:
            self.ctx.clock.sleep(0.001)
            if self.ctx.clock.monotonic() > timeout:
                return False
        return True

    def _switch_channel(self, channel: Literal[1, 2]) -> bool:
        if not self.ctx:
            return False
        if self._current_channel == channel:
            return True

        self.device.set_channel(channel)

        # Discard 10 samples after changing channels
        for n in range(1, 10):
            if not self._wait_ready():
                self.ctx.emit(
                    DeviceFailed(
                        device=self.id,
                        kind=self.kind,
                        error="timeout changing channels",
                    )
                )
                return False
            self.device.adco

        self._current_channel = channel
        return True

    def read(self, channel: Literal[1, 2]) -> int | None:
        if not self._switch_channel(channel):
            self.set_offline()
            return None

        with self._lock:
            try:
                value = self.device.adco.value
            except Exception as e:
                self.set_offline(e)
                return None

        self.set_online()
        return value



from typing import Optional, Any

import RPi.GPIO as GPIO

from bussdcc.device import Device

from .config import PumpConfig


class Pump(Device[PumpConfig]):
    kind = "actuator"

    def __init__(self, *, id: str, config: PumpConfig):
        super().__init__(id=id, config=config)
        self._state = False

    def connect(self) -> None:
        GPIO.setmode(GPIO.BCM)
        GPIO.setup(self.config.pin, GPIO.OUT)
        GPIO.output(self.config.pin, GPIO.LOW)

    def on(self) -> None:
        GPIO.output(self.config.pin, GPIO.HIGH)
        self._state = True

    def off(self) -> None:
        GPIO.output(self.config.pin, GPIO.LOW)
        self._state = False

    def disconnect(self) -> None:
        self.off()
        GPIO.cleanup(self.config.pin)
